#include <EEPROM.h>

#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))

int pwm,cc,dd;

float Vsp, Isp, hs_temp;
float Vin,Vout,Iin,Iout;
volatile float xVin,xVout,xIin,xIout,tIin, tIout;
int oIin,oIout;


void get_biquad(float,int,int);
void biquad_filter(float *,int);

float aa[5][4],d1[5][4],d2[5][4];
float w0[5][4],w1[5][4],w2[5][4];
int bqnp[4];

volatile int f50c,f50,octrip;

volatile char buf[80],*bp,bcount;
volatile uint8_t blcd,go_lcd = 0,bpos=0;
volatile int baudc = 0,freq_scale;
unsigned int ints_second;
int baud_waits,timer1_period;


void get_biquad(float fc, int np, int idx)
  {  
  float s,r;
  int i;  
  float a = tan(M_PI*fc);
  float a2 = a*a;
  bqnp[idx] = np;
  for(i=0; i<np; ++i)
    {
    r =  sin(3.141592926 * (2.0*i+1.0)/(4.0*(float)np));
    s = a2 + 2.0*a*r + 1.0;
    aa[i][idx] = a2/s;
    d1[i][idx] = 2.0*(1-a2)/s;
    d2[i][idx] = -(a2 - 2.0*a*r + 1.0)/s;
    w1[i][idx]=w2[i][idx]=w0[i][idx]=0.0;
    }
  }

void biquad_filter(float *val, int idx)
{ 
  float s;
  int i;
  s = *val;
  for(i=0; i<bqnp[idx]; ++i)
    {
    w0[i][idx] = d1[i][idx]*w1[i][idx] + d2[i][idx]*w2[i][idx] + s;
    s = aa[i][idx]*(w0[i][idx] + 2.0*w1[i][idx] + w2[i][idx]);
    w2[i][idx] = w1[i][idx];
    w1[i][idx] = w0[i][idx];
    }
  *val = s;
}

float get_temp(int Vo,int hi_low)
{
 float a = 1.027957148e-3, b = 2.393569459e-4, c = 1.555071887e-7;
 float T,logRt,Tc,v,r,i;                                 // assumes NTC on high side of voltage divider
 v = 5.0 * (float)Vo /1024.0;    // v in volts
 if (hi_low == 1)     // NTC is on high side of divider, with 10K on low side
  {
    i = v/1e4;
    r = (5.0 - v)/i;
  }
 else                 // NTC is on low side of divider with 10K on high side
  {
    i = (5.0 - v)/1e4;
    r = v/i;
  }
 logRt = log(r); 
 T = (1.0 / (a + b*logRt + c*logRt*logRt*logRt)); // We get the temperature value in Kelvin from this Stein-Hart equation
 Tc = T - 273.15;                                 // Convert Kelvin to Celsius
 return Tc;
}



void setup()
  {  
  pwm = 1;
  pinMode(7,INPUT);   // let comparator inputs float, dont drive them internally
  pinMode(6,INPUT);

  ACSR  |=  (1<<ACI);    // clear Analog Comparator interrupt
  ACSR  |=
        (0<<ACD)   |         // Comparator ON
        (0<<ACBG)  |         // Disconnect 1.23V reference from AIN0 (use AIN0 and AIN1 pins)
        (0<<ACIE)  |         // Comparator Interrupt enabled
        (0<<ACIC)  |         // input capture disabled
        (0<<ACIS1) |         // set interrupt bit on rising edge
        (0<<ACIS0);          //    (ACIS1 and ACIS0 == 11)
  
   pinMode(18,OUTPUT);         // my software serial port, 9600 baud, uses pin A4
  bp = &buf[0];
  for(bcount=0; bcount < 80; bcount++)
    buf[bcount]=20;
  bcount = 0;

  timer1_period = 833;
    baud_waits = 1;
    ints_second = 19200;
    freq_scale = 1;
    
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);
  TCCR1A = _BV(COM1A1) | _BV(COM1B1) | _BV(WGM11);
  TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS10);
  OCR1A = 1;
  //OCR1B = 2;
  ICR1 = timer1_period;
  TIMSK1 |= (1 << TOIE1);   // enable timer overflow interrupt
  Serial.begin(115200);
  
  Vsp = 10.0;
  Isp = 55.0; 

  get_biquad(0.1,2,0);
  get_biquad(0.1,2,1);
  get_biquad(0.02,2,2);
  get_biquad(0.02,2,3);
  cc = 0;
  dd = 0;
  
  f50c=0;
  f50=0;
xVin = 0.1744;
  xVout = 0.0900;
  xIin = 0.235;
  xIout = 0.244;
  tIin = 0.00;
  tIout = 0.00;
  oIin = 511;
  oIout = 511;
  octrip = 0;
  Serial.println("setup complete");
 }

ISR(TIMER1_OVF_vect)        
{
if (0 && (ACSR & 0x10) == 0x10 )      // no over current testing, it just gets in the way sometimes
  {       
    OCR1A = 1;
    pwm = 1;
    ACSR = 0; 
    octrip = 1;
  }
  
OCR1A = pwm;    // pin 9

// send one char out pin A4. First, get a character and put it in "blcd" and init state machine.
if (go_lcd == 0 && baudc > 50)    // not send as soon as possible, let the LCD digest the data with a delay here
  {
  // get next char
  blcd = buf[bcount];
  bcount++;
  if (bcount > 79)
    {
    bcount = 0;
    blcd =  '$';     // this is translated by the serial LCD as "put cursor at top left position"
    }   
  go_lcd = 1;
  bpos = 0;
  baudc = 2;
  }
  
if (go_lcd == 1 && baudc > baud_waits)   // a little 8 bit serial output state machine
  {
    if(bpos == 9)
    {
    // send stop bit, pull high
    sbi(PORTC,4);
    go_lcd = 0;
    baudc = 0;
    bpos = 0;
    }
  if (bpos > 0  )
    {
      if (blcd & 1)
        sbi(PORTC,4);
      else
        cbi(PORTC,4);
      blcd = blcd >> 1;
      bpos++;  
      baudc = 0;
    }
  if(bpos == 0 && go_lcd == 1)
    {
    // send start bit, pull low
    cbi(PORTC,4);
    bpos++;  
    baudc = 0;
    }
  }
  baudc++;              // a counter, goes from 0, 1, 0 , 1 etc. to halve the output frequency of the serial data
  
f50c++;
if(f50c > 99)
  {
    f50c=0;
    f50=1;
  }
}   




   
void loop()
   {
   int a0,fa,fi,nupwm;
   float p,e;
   char fbuf[10],gbuf[10],hbuf[10];
   
   hs_temp = get_temp(analogRead(7),0);
   
   // voltage setpoint via pot on A5, 0 to 45V 
   Vsp = Vsp * 0.99 + 0.01 * 75.0 * analogRead(5)/1024.0; 
      
   // get output voltage   
   Vout = xVout * (float)analogRead(2);
   //biquad_filter(&Vout,0); 

   // get input voltage
   Vin = xVin * (float)analogRead(0);
   biquad_filter(&Vin,1); 

   // get output current
   Iout = - tIout + xIout * (float)(analogRead(3) - oIout);
   biquad_filter(&Iout,2); 

   // get input current
   Iin = -tIin + xIin * (float)(analogRead(1) - oIin);
   biquad_filter(&Iin,3); 
    
   if (f50 == 1)
    {
    f50=0;
    nupwm = pwm;
    if ( Vin < Vsp)    // if input V is less than output V, shut down    
          nupwm -= 5;                
    
      if (Vout < Vsp && Iout < Isp && ( Vin > Vsp))   nupwm ++;  
      if (Vout > Vsp || Iout > Isp || ( Vin < Vsp))   nupwm --;  
      if (Iout > Isp)       nupwm -= 5;    
      
      // clamp to reasonable values
      if (nupwm > 799) nupwm = 799;
      if (nupwm < 1) nupwm = 1;
      pwm = nupwm;
      

    }

  
   
   if (dd++ > 200)
    {
      dd=0; 

      Serial.print(Vout * Iout /10.0);      Serial.print(" "); 
      Serial.print(Vin * Iin /10.0);      Serial.print(" "); 
      Serial.print(Vin);      Serial.print(" "); 
      Serial.print(Vout);      Serial.print(" ");  
      Serial.print(Iin);      Serial.print(" "); 
      Serial.print(Iout);      Serial.print(" ");  
      Serial.print(Vsp);      Serial.print(" "); 
      Serial.print(" 200 "); 
      Serial.println();

      /*
      for (a0=0; a0 < 8; a0++)
        {
          Serial.print(analogRead(a0)); Serial.print (" ");
        }
      Serial.println();
      */  
      dtostrf(Vin,4,1,fbuf);  
      dtostrf(Iin,4,1,hbuf); 
      dtostrf(hs_temp,3,0,gbuf);      
      sprintf(bp,"In  %4sV %4sA %3sC",fbuf,hbuf,gbuf); 
      dtostrf(Vout,4,1,fbuf);  
      dtostrf(Iout,4,1,hbuf);       
      sprintf(bp+40,"Out %4sV %4sA    ",fbuf,hbuf); buf[59] = 20; 
      p = Vout*Iout;
      dtostrf(p,5,0,fbuf);        
      sprintf(bp+20,"%5sW Out         ",fbuf); buf[39] = 20;
      dtostrf(Vsp,4,1,fbuf);        
      sprintf(bp+60,"setpoint %4s      ",fbuf); buf[59] = 20; 
      if (octrip == 1)
        {
          sprintf(bp+60,"OC tripped          "); buf[59] = 20; 
        }
      else
        {
         dtostrf(Vsp,4,1,fbuf);        
        sprintf(bp+60,"setpoint %4s      ",fbuf); buf[59] = 20;  
        }
    }  
   }

 
